#include <infiniband/verbs.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <pthread.h>
#include <stdarg.h>
#include <sys/time.h>

#define pr_debug(level, fmt, ...)		\
do {						\
	if (verbs_level >= level)		\
		printf(fmt, ##__VA_ARGS__);	\
} while(0)

#define error_out(fmt, ...)			\
do {						\
	pr_debug(1, fmt, ##__VA_ARGS__);	\
	pr_debug(1, "Reason: %s\n", strerror(errno));	\
	goto out;				\
} while(0)

#define pr_info(fmt, ...)			\
do {						\
	pr_debug(2, fmt, ##__VA_ARGS__);	\
} while(0)

#define pr_detail(fmt, ...)			\
do {						\
	pr_debug(3, fmt, ##__VA_ARGS__);	\
} while(0)

enum {
	TYPE_WRITE,
	TYPE_IMM,
	TYPE_RECV,
	TYPE_STATE,
};

enum {
	STATE_PREPARE,
	STATE_READY,
	STATE_END,
};

static struct ibv_pd *pd;
static struct ibv_cq *cqs[2];
static struct ibv_qp *qps[2];
static struct ibv_mr *mr;
static struct ibv_context *ctx;
static struct ibv_device_attr dev_attr;
static struct ibv_qp_init_attr qp_iattr;
static struct ibv_qp_attr qp_attr;
static struct ibv_port_attr port_attr;
static union ibv_gid gid;

static unsigned int verbs_level = 2;
static unsigned int gid_index = 1;
static unsigned int dev_index = 0;
static unsigned int port_index = 1;
static unsigned int is_master = 1;
static unsigned int master_port = 12345;
static unsigned int nr_round = 1;
static unsigned int delay_us = 1000;
static unsigned int max_requests = 10;
static unsigned int request_size = 4096000;
static unsigned int buffer_size;

static int lfs, cfd;
static char master_ip[64];
static struct qp_info lqpi, rqpi;
static char *buffer;
static int *send_state, *recv_state;
unsigned long long cost_total = 0;
unsigned long long cost_real = 0;
unsigned long long cost_start = 0;
unsigned long long cost_last = 0;
unsigned long long send_total = 0;
unsigned long long send_last = 0;

static pthread_t worker_task;

struct qp_info {
	int ready;
	unsigned int lid;
	unsigned int qpn[2];
	unsigned int rkey;
	enum ibv_mtu active_mtu;
	unsigned long long subnet_prefix;
	unsigned long long interface_id;
	unsigned long long addr;
};

struct request {
	int type;
	int state;
	char *buf;
	size_t size;
	unsigned long long ts;
} *requests;

static unsigned long long ts_us(void)
{
	struct timeval tv;

	if (gettimeofday(&tv, NULL)) {
		pr_info("Get time failed\n");
		return 0;
	}

	return tv.tv_sec * 1000000 + tv.tv_usec;
}

static int hand_shake(void)
{
	if (is_master) {
		pr_detail("Wait info...\n");

		if (recv(cfd, &nr_round, sizeof(int), 0) <= 0)
			return -EINVAL;

		pr_detail("Send ready...\n");

		if (send(cfd, &nr_round, sizeof(int), 0) <= 0)
			return -EINVAL;

		sleep(1);
	} else {
		pr_detail("Send info...\n");

		if (send(cfd, &nr_round, sizeof(int), 0) <= 0)
			return -EINVAL;

		pr_detail("Wait ready...\n");

		if (recv(cfd, &nr_round, sizeof(int), 0) <= 0)
			return -EINVAL;
	}

	return 0;
}

static int register_mr(void *buf, size_t size)
{
	mr = ibv_reg_mr(pd, buf, size, IBV_ACCESS_REMOTE_WRITE|IBV_ACCESS_LOCAL_WRITE|IBV_ACCESS_REMOTE_READ);
	if (!mr)
		return -EINVAL;

	return 0;
}

static int modify_qp(struct qp_info *qpi)
{
	int i;

	for (i = 0; i < 2; i++) {
		memset(&qp_attr, 0, sizeof(struct ibv_qp_attr));
		qp_attr.qp_state = IBV_QPS_INIT;
		qp_attr.pkey_index = 0;
		qp_attr.port_num = 1;
		qp_attr.qp_access_flags = IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ;

		if (ibv_modify_qp(qps[i], &qp_attr, IBV_QP_STATE | IBV_QP_PKEY_INDEX | IBV_QP_PORT | IBV_QP_ACCESS_FLAGS))
			return -EINVAL;

		pr_detail("QP %d inited\n", qps[i]->qp_num);

		memset(&qp_attr, 0, sizeof(struct ibv_qp_attr));
		qp_attr.qp_state = IBV_QPS_RTR;
		qp_attr.path_mtu = qpi->active_mtu;
		qp_attr.dest_qp_num = qpi->qpn[i];
		qp_attr.rq_psn = 0;
		qp_attr.max_dest_rd_atomic = 1;
		qp_attr.min_rnr_timer = 12;

		qp_attr.ah_attr.is_global = 0;
		qp_attr.ah_attr.dlid = qpi->lid;
		qp_attr.ah_attr.sl = 0;
		qp_attr.ah_attr.src_path_bits = 0;
		qp_attr.ah_attr.port_num = 1;

		qp_attr.ah_attr.is_global = 1;
		qp_attr.ah_attr.grh.dgid.global.subnet_prefix = qpi->subnet_prefix;
		qp_attr.ah_attr.grh.dgid.global.interface_id = qpi->interface_id;
		qp_attr.ah_attr.grh.flow_label = 0;
		qp_attr.ah_attr.grh.sgid_index = gid_index;
		qp_attr.ah_attr.grh.hop_limit = 255;
		qp_attr.ah_attr.grh.traffic_class = 0;

		if (ibv_modify_qp(qps[i], &qp_attr, IBV_QP_STATE | IBV_QP_AV | IBV_QP_PATH_MTU | IBV_QP_DEST_QPN | IBV_QP_RQ_PSN | IBV_QP_MAX_DEST_RD_ATOMIC | IBV_QP_MIN_RNR_TIMER))
			return -EINVAL;

		pr_detail("QP %d ready to receive\n", qps[i]->qp_num);

		memset(&qp_attr, 0, sizeof(struct ibv_qp_attr));
		qp_attr.qp_state = IBV_QPS_RTS;
		qp_attr.timeout = 18;
		qp_attr.retry_cnt = 7;
		qp_attr.rnr_retry = 7;
		qp_attr.sq_psn = 0;
		qp_attr.max_rd_atomic = 1;
		if (ibv_modify_qp(qps[i], &qp_attr, IBV_QP_STATE | IBV_QP_TIMEOUT | IBV_QP_RETRY_CNT | IBV_QP_RNR_RETRY | IBV_QP_SQ_PSN | IBV_QP_MAX_QP_RD_ATOMIC))
			return -EINVAL;

		pr_detail("QP %d ready to send\n", qps[i]->qp_num);
	}

	return 0;
}

static int create_qp(void)
{
	int i;

	pd = ibv_alloc_pd(ctx);
	if (!pd)
		return -EINVAL;

	for (i = 0; i < 2; i++) {
		cqs[i] = ibv_create_cq(ctx, max_requests, NULL, NULL, 0);
		if (!cqs[i])
			return -EINVAL;

		memset(&qp_iattr, 0, sizeof(struct ibv_qp_init_attr));
		qp_iattr.send_cq = cqs[i];
		qp_iattr.recv_cq = cqs[i];
		qp_iattr.qp_type = IBV_QPT_RC;
		qp_iattr.cap.max_send_wr = max_requests;
		qp_iattr.cap.max_recv_wr = max_requests;
		qp_iattr.cap.max_send_sge = 1;
		qp_iattr.cap.max_recv_sge = 1;
		qp_iattr.cap.max_inline_data = 0;

		qps[i] = ibv_create_qp(pd, &qp_iattr);
		if (!qps[i])
			return -EINVAL;

		pr_detail("QP %d created\n", qps[i]->qp_num);
	}

	return 0;
}

static int send_qpi(int fd, void *buf, struct qp_info *qpi)
{
	memset(qpi, 0, sizeof(*qpi));
	qpi->ready = 1;
	qpi->lid = port_attr.lid;
	qpi->qpn[0] = qps[0]->qp_num;
	qpi->qpn[1] = qps[1]->qp_num;
	qpi->active_mtu = port_attr.active_mtu;
	qpi->subnet_prefix = gid.global.subnet_prefix;
	qpi->interface_id = gid.global.interface_id;
	qpi->rkey = mr->rkey;
	qpi->addr = (unsigned long long)buf;

	if (send(fd, qpi, sizeof(*qpi), 0) <= 0)
		return -EINVAL;

	pr_detail("Sent QP info:\n");
	pr_detail("\tqpn0 = %d\n", qpi->qpn[0]);
	pr_detail("\tqpn1 = %d\n", qpi->qpn[1]);
	pr_detail("\tlid = %d\n", qpi->lid);
	pr_detail("\trkey = %d\n", qpi->rkey);
	pr_detail("\taddr = %llx\n", qpi->addr);
	pr_detail("\tactive_mtu = %d\n", qpi->active_mtu);
	pr_detail("\tsubnet_prefix = %llx\n", qpi->subnet_prefix);
	pr_detail("\tinterface_id = %llx\n", qpi->interface_id);

	return 0;
}

static int recv_qpi(int fd, struct qp_info *qpi)
{
	memset(qpi, 0, sizeof(*qpi));
	if (recv(fd, qpi, sizeof(*qpi), 0) <= 0)
		return -EINVAL;

	pr_detail("Received QP info:\n");
	pr_detail("\tqpn0 = %d\n", qpi->qpn[0]);
	pr_detail("\tqpn1 = %d\n", qpi->qpn[1]);
	pr_detail("\tlid = %d\n", qpi->lid);
	pr_detail("\trkey = %d\n", qpi->rkey);
	pr_detail("\taddr = %llx\n", qpi->addr);
	pr_detail("\tactive_mtu = %d\n", qpi->active_mtu);
	pr_detail("\tsubnet_prefix = %llx\n", qpi->subnet_prefix);
	pr_detail("\tinterface_id = %llx\n", qpi->interface_id);

	return 0;
}

static int poll_cq(int idx)
{
	int i, done;
	struct request *req;
	struct ibv_wc wcs[4];

	done = ibv_poll_cq(cqs[idx], 4, wcs);
	if (done < 0)
		error_out("Poll CQ failed\n");

	for (i = 0; i < done; i++) {
		struct ibv_wc *wc = &wcs[i];
		int id = wc->wr_id;
		struct request *req = &requests[id];

		switch (req->type) {
		case TYPE_WRITE:
		case TYPE_IMM:
		case TYPE_STATE:
			if (wc->status != IBV_WC_SUCCESS)
				error_out("REQ %d send failed\n", id);
			else
				pr_detail("REQ %d send succeed\n", id);

			if (req->type == TYPE_STATE)
				break;

			req->ts = ts_us() - req->ts;
			pr_detail("REQ %d size %lu cost %llu\n", id, req->size, req->ts);
			if (req->type == TYPE_IMM) {
				send_total += req->size * (max_requests - 2);
				cost_total += ts_us() - cost_start;
			}

			break;
		case TYPE_RECV:
			if (wc->status != IBV_WC_SUCCESS)
				error_out("REQ %d recv failed\n", id);
			else
				pr_detail("REQ %d recv succeed\n", id);
			if (wc->imm_data == STATE_END) {
				pr_detail("Last recv end\n");
				*recv_state = STATE_END;
			} else
				*recv_state = STATE_PREPARE;

			break;
		default:
			pr_detail("REQ %d unknow op=%d\n", id, wc->opcode);
		};
	}

	return done;
out:
	return -EINVAL;
}

static int do_send(int round)
{
	int i;
	struct ibv_sge sge[max_requests];
	struct ibv_send_wr swr[max_requests], *bad_swr;

	for (i = 2; i < max_requests; i++) {
		struct request *req = &requests[i];

		req->ts = ts_us();
		req->size = request_size;

		memset(&sge[i], 0, sizeof(struct ibv_sge));
		memset(&swr[i], 0, sizeof(struct ibv_send_wr));

		sge[i].addr = (unsigned long)req->buf;
		sge[i].length = request_size;
		sge[i].lkey = mr->lkey;

		swr[i].wr_id = i;
		if (i == max_requests - 1) {
			swr[i].opcode = IBV_WR_RDMA_WRITE_WITH_IMM;
			req->type = TYPE_IMM;
			swr[i].send_flags = IBV_SEND_SIGNALED;
			swr[i].next = NULL;
		} else {
			swr[i].opcode = IBV_WR_RDMA_WRITE;
			req->type = TYPE_WRITE;
			swr[i].send_flags = 0;
			swr[i].next = &swr[i + 1];
		}
		swr[i].wr.rdma.rkey = rqpi.rkey;
		swr[i].wr.rdma.remote_addr = rqpi.addr + sizeof(int) * 2 + i * request_size;
		swr[i].sg_list = &sge[i];
		swr[i].num_sge = 1;

		if (round >= nr_round)
			swr[i].imm_data = STATE_END;
	}

	if (ibv_post_send(qps[!!is_master], &swr[2], &bad_swr))
		error_out("Send failed\n");

	pr_detail("REQ send post\n");

	return 0;
out:
	return -EINVAL;
}

static int do_recv(void)
{
	int i;
	struct ibv_sge sge;
	struct ibv_send_wr swr, *bad_swr;
	struct ibv_recv_wr rwr, *bad_rwr;

	memset(&rwr, 0, sizeof(rwr));

	requests[0].type = TYPE_RECV;
	requests[0].ts = ts_us();
	rwr.wr_id = 0;
	rwr.sg_list = NULL;
	rwr.num_sge = 0;

	if (ibv_post_recv(qps[!is_master], &rwr, &bad_rwr))
		error_out("Receive failed\n");

	pr_detail("REQ %d recv post\n", 3);

	memset(&sge, 0, sizeof(struct ibv_sge));
	memset(&swr, 0, sizeof(struct ibv_send_wr));

	requests[1].type = TYPE_STATE;
	requests[1].ts = ts_us();
	requests[1].size = sizeof(int);

	sge.addr = (unsigned long)recv_state;
	sge.length = sizeof(int);
	sge.lkey = mr->lkey;

	swr.wr_id = 1;
	swr.next = NULL;
	swr.opcode = IBV_WR_RDMA_WRITE;
	swr.send_flags = IBV_SEND_SIGNALED;
	swr.wr.rdma.rkey = rqpi.rkey;
	swr.wr.rdma.remote_addr = rqpi.addr;
	swr.sg_list = &sge;
	swr.num_sge = 1;

	if (ibv_post_send(qps[!is_master], &swr, &bad_swr))
		error_out("Send failed\n");

	pr_detail("REQ 0 send post\n");

	return 0;
out:
	return -EINVAL;
}

static void *worker_func(void *arg)
{
	int r = 0, skip_send = 0;
	struct request *req;

	while (*recv_state != STATE_END || !skip_send) {
		if (poll_cq(0) < 0 || poll_cq(1) < 0)
			break;

		if (*recv_state == STATE_PREPARE) {
			*recv_state = STATE_READY;
			if (do_recv())
				break;
		}

		if (!skip_send && *send_state == STATE_READY) {
			if (r)
				pr_info("Round %d send %llu cost %llu real %llu\n", r, send_total - send_last, cost_total - cost_last, ts_us() - cost_real);
			else
				cost_real = ts_us();

			usleep(delay_us);
			if (r >= nr_round)
				skip_send = 1;

			send_last = send_total;
			cost_last = cost_total;
			cost_start = ts_us();

			if (do_send(r++))
				break;

			*send_state = STATE_PREPARE;
			if (skip_send)
				pr_detail("Last send end\n");
		}
	};

	cost_real = ts_us() - cost_real;
	pr_debug(0, "Summary:\n"			\
		"  send %0.2fGb in %0.2fsec total %0.2f sec\n"	\
		"  rdma_bw %0.2fGB/s real_bw %0.2fGB/s\n",
			(float)send_total / 1073741824,
			(float)cost_total / 1000000,
			(float)cost_real / 1000000,
			(float)(send_total * 8000000) / (float)(cost_total * 1073741824),
			(float)(send_total * 8000000) / (float)(cost_real * 1073741824));

out:
	return NULL;
}

static int master_start(void)
{
	int ret = -EINVAL;
	socklen_t socklen = sizeof(struct sockaddr_in);
	struct sockaddr_in saddr, caddr;

	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr = INADDR_ANY;
	saddr.sin_port = htons(master_port);

	lfs = socket(AF_INET, SOCK_STREAM, 0);
	if (lfs <= 0)
		error_out("Create socket failed\n");

	if (bind(lfs, (struct sockaddr *)&saddr, socklen))
		error_out("Bind failed\n");

	if (listen(lfs, 8))
		error_out("Listen failed\n");

	pr_info("Master accepting\n");
	cfd = accept(lfs, (struct sockaddr *)&caddr, &socklen);
	if (cfd <= 0)
		error_out("Accept failed\n");

	if (create_qp())
		error_out("Create QP failed\n");

	if (register_mr(buffer, buffer_size))
		error_out("Register MR failed\n");

	if (recv_qpi(cfd, &rqpi))
		error_out("Receive QP info failed\n");

	if (modify_qp(&rqpi))
		error_out("Modify QP failed\n");

	if (send_qpi(cfd, buffer, &lqpi))
		error_out("Send QP info failed\n");

	if (hand_shake())
		error_out("Wait Slave failed\n");

	if (pthread_create(&worker_task, NULL, worker_func, NULL))
		error_out("Create pthread failed\n");

	pr_info("Master ready\n");

	ret = 0;
out:
	return ret;
}

static int slave_start()
{
	int ret = -EINVAL;
	struct hostent *hostnm;
	struct sockaddr_in saddr;

	hostnm = gethostbyname(master_ip);
	if (!hostnm)
		error_out("Get host failed\n");

	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr = *((unsigned long *)hostnm->h_addr);
	saddr.sin_port = htons(master_port);

	cfd = socket(AF_INET, SOCK_STREAM, 0);
	if (cfd <= 0)
		error_out("Create socket failed\n");

	pr_info("Slave connecting\n");
	if (connect(cfd, (struct sockaddr *)&saddr, sizeof(saddr)))
		error_out("Connect failed\n");

	if (create_qp())
		error_out("Create QP failed\n");

	if (register_mr(buffer, buffer_size))
		error_out("Register MR failed\n");

	if (send_qpi(cfd, buffer, &lqpi))
		error_out("Send QP info failed\n");

	if (recv_qpi(cfd, &rqpi))
		error_out("Receive QP info failed\n");

	if (modify_qp(&rqpi))
		error_out("Modify QP failed\n");

	if (hand_shake())
		error_out("Wakeup master failed\n");

	if (pthread_create(&worker_task, NULL, worker_func, NULL))
		return -EINVAL;

	pr_info("Slave ready\n");

	ret = 0;
out:
	return ret;
}

int main(int argc, char **argv)
{
	int i, opt, num_devs;
	struct ibv_device **devs = NULL, *dev;

	while ((opt = getopt(argc, argv, "m:p:v:r:g:d:i:s:")) != -1) {
		switch (opt) {
		case 'm':
			memcpy(master_ip, optarg, sizeof(master_ip));
			is_master = 0;
			break;
		case 'p':
			master_port = atoi(optarg);
			break;
		case 'v':
			verbs_level = atoi(optarg);
			break;
		case 'r':
			nr_round = atoi(optarg);
			break;
		case 'g':
			gid_index = atoi(optarg);
			break;
		case 'd':
			delay_us = atoi(optarg);
			break;
		case 'i':
			dev_index = atoi(optarg);
			break;
		case 's':
			sscanf(optarg, "%u:%u", &max_requests, &request_size);
			max_requests += 2;
			break;
		default:
			error_out("Usage: %s [-v VERBS] [-r ROUND] [-g GID] [-m MASTER] [-p PORT] [-d DELAY] [-s NR_REQUEST:REQUEST_SIZE]\n", argv[0]);
		}
	}

	if (max_requests < 3)
		error_out("Request number is zero\n");

	buffer_size = max_requests * request_size + sizeof(int) * 2;
	buffer = malloc(buffer_size);
	if (!buffer)
		error_out("Out of memory\n");

	requests = malloc(max_requests * sizeof(struct request));
	if (!requests)
		error_out("Out of memory\n");

	send_state = (int *)buffer;
	recv_state = send_state + 1;
	*recv_state = *send_state = STATE_PREPARE;
	for (i = 0; i < max_requests; i++) {
		requests[i].type = -1;
		requests[i].buf = buffer + sizeof(int) * 2 + i * request_size;
	}

	if (is_master)
		pr_info("This is a master\n");
	else
		pr_info("This is a slave, master is %s\n", master_ip);

	devs = ibv_get_device_list(&num_devs);
	if (!devs || !num_devs)
		error_out("Get IB devices failed\n");

	for (i = 0; i < num_devs; i++)
		pr_detail("find IB device %s\n", devs[i]->name);

	dev = devs[dev_index];
	ctx = ibv_open_device(dev);
	if (!ctx)
		error_out("Open device failed\n");

	if (ibv_query_device(ctx, &dev_attr))
		error_out("Query device failed\n");

	if (ibv_query_port(ctx, port_index, &port_attr))
		error_out("Query port failed\n");

	if (ibv_query_gid(ctx, port_index, gid_index, &gid))
		error_out("Query gid failed\n");

	pr_info("Using IB device %s port %d:\n", dev->name, port_index);
	pr_info("\tlink_layer = %d\n", port_attr.link_layer);
	pr_info("\tactive_mtu = %d\n", port_attr.active_mtu);
	pr_info("\tlid = %d\n", port_attr.lid);
	pr_info("\tgid%d_subnet_prefix = %llx\n", gid_index, gid.global.subnet_prefix);
	pr_info("\tgid%d_interface_id = %llx\n", gid_index, gid.global.interface_id);

	if (is_master) {
		if (master_start())
			goto out;
	} else {
		if (slave_start())
			goto out;
	}

	pthread_join(worker_task, NULL);
	if (hand_shake())
		goto out;

	pr_info("Finished\n");

out:
	if (cfd > 0)
		close(cfd);
	if (lfs > 0)
		close(lfs);
	if (mr)
		ibv_dereg_mr(mr);
	for (i = 0; i < 2; i++) {
		if (qps[i])
			ibv_destroy_qp(qps[i]);
		if (cqs[i])
			ibv_destroy_cq(cqs[i]);
	}
	if (pd)
		ibv_dealloc_pd(pd);
	if (ctx)
		ibv_close_device(ctx);
	if (devs)
		ibv_free_device_list(devs);
	if (buffer)
		free(buffer);
	return 0;
}
