#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/sock_diag.h>
#include <linux/unix_diag.h>
#include <linux/inet_diag.h>
#include <linux/in.h>
#include <linux/version.h>
#include <linux/tcp.h>
#include <dirent.h>
#include <time.h>

typedef unsigned long long u64;

#define CC(cond, fmt, ...)		\
do {						\
	if (cond) {				\
		printf(fmt, ##__VA_ARGS__);	\
		return -1;			\
	}					\
} while (0)

#define KERNEL_VERSION_AT_LEAST(v, u) \
    (LINUX_VERSION_CODE >= KERNEL_VERSION(v, u, 0))

#define IP_SIZE		32
#define PATH_SIZE	64
#define PRINT_SIZE	64
#define COMM_SIZE	128
#define MSG_SIZE	512

#define COMM_MODE	0x1
#define NOCOMM_MODE	0x2
#define SRC_MODE	0x4
#define NOSRC_MODE	0x8
#define DST_MODE	0x10
#define NODST_MODE	0x20
#define LOG_MODE	0x40
#define TOP_MODE	0x80
#define FILTER_MODE	0x100

static char log_path[PATH_SIZE];
static FILE *log_fd;
static int mode;
static int order;
static int top_show;
static int rr_wmark;
static u64 recv_wmark;
static u64 sent_wmark;

enum {
	RR_ORDER,
	RECV_ORDER,
	SENT_ORDER,
};

enum {
	TCP_ESTABLISHED = 1,
	TCP_SYN_SENT,
	TCP_SYN_RECV,
	TCP_FIN_WAIT1,
	TCP_FIN_WAIT2,
	TCP_TIME_WAIT,
	TCP_CLOSE,
	TCP_CLOSE_WAIT,
	TCP_LAST_ACK,
	TCP_LISTEN,
	TCP_CLOSING,
	TCP_NEW_SYN_RECV,

	TCP_MAX_STATES
};

char *state_name[TCP_MAX_STATES] = {
	"N/A",
	"ESTABLISHED",
	"SYN_SENT",
	"SYN_RECV",
	"FIN_WAIT1",
	"FIN_WAIT2",
	"TIME_WAIT",
	"CLOSE",
	"CLOSE_WAIT",
	"LAST_ACK",
	"LISTEN",
	"CLOSING",
	"NEW_SYN_RECV",
};

char *ca_state_name[5] = {
	"NORMAL",
	"DISORDER",
	"CONGESTION",
	"RECOVER",
	"LOSS",
};

static int monitor_states =
	1<<TCP_ESTABLISHED	|
	1<<TCP_SYN_SENT		|
	1<<TCP_SYN_RECV		|
	1<<TCP_FIN_WAIT1	|
	1<<TCP_FIN_WAIT2	|
	1<<TCP_TIME_WAIT	|
	1<<TCP_CLOSE		|
	1<<TCP_CLOSE_WAIT	|
	1<<TCP_LAST_ACK		|
//	1<<TCP_LISTEN		|
	1<<TCP_CLOSING		|
	1<<TCP_NEW_SYN_RECV;

#define HASH_SIZE 512

struct hash_node {
	int seq;
	int pid;
	int inode;
	int first;
	int disable;
	int src_port;
	int dst_port;
	float rr;
	struct tcp_info ti;
	struct tcp_info diff;
	char comm[COMM_SIZE];
	char src_ip[IP_SIZE];
	char dst_ip[IP_SIZE];
	struct hash_node *next;
	struct hash_node *plist;
};

static char comm_filter[COMM_SIZE];
static char nocomm_filter[COMM_SIZE];
static char src_filter[IP_SIZE];
static char nosrc_filter[IP_SIZE];
static char dst_filter[IP_SIZE];
static char nodst_filter[IP_SIZE];

static int setup_target_ip(char *optarg)
{
	if (sscanf(optarg, "src!=%s", nosrc_filter) == 1)
		mode |= NOSRC_MODE;
	else if (sscanf(optarg, "src=%s", src_filter) == 1)
		mode |= SRC_MODE;
	else if (sscanf(optarg, "dst!=%s", nodst_filter) == 1)
		mode |= NODST_MODE;
	else if (sscanf(optarg, "dst=%s", dst_filter) == 1)
		mode |= DST_MODE;
	else
		return 1;
	return 0;
}
static int target_ip(struct hash_node *hn)
{
	if ((mode & DST_MODE) && strcmp(hn->dst_ip, dst_filter))
		return 0;
	if ((mode & NODST_MODE) && !strcmp(hn->dst_ip, nodst_filter))
		return 0;
	if ((mode & SRC_MODE) && strcmp(hn->src_ip, src_filter))
		return 0;
	if ((mode & NOSRC_MODE) && !strcmp(hn->src_ip, nosrc_filter))
		return 0;
	return 1;
}

static int setup_target_comm(char *optarg)
{
	if (sscanf(optarg, "comm=%s", comm_filter) == 1)
		mode |= COMM_MODE;
	else if (sscanf(optarg, "comm!=%s", nocomm_filter) == 1)
		mode |= NOCOMM_MODE;
	else
		return 1;
	return 0;
}

static int target_comm(char *comm)
{
	if ((mode & COMM_MODE) && strcmp(comm, comm_filter))
		return 0;
	if ((mode & NOCOMM_MODE) && !strcmp(comm, nocomm_filter))
		return 0;
	return 1;
}

static struct hash_node *hash_lists[HASH_SIZE];
static struct hash_node *show_list;
static int nr_show;

static struct hash_node *alloc_hash_node(int inode, int pid, char *comm, int seq)
{
	struct hash_node *hn;

	hn = malloc(sizeof(struct hash_node));
	if (hn) {
		hn->disable = 0;
		hn->first = 1;
		hn->seq = seq;
		hn->pid = pid;
		hn->inode = inode;
		memcpy(hn->comm, comm, COMM_SIZE);
		hn->next = NULL;
	}

	return hn;
}

static int inline get_hash(int inode)
{
	return ((inode >> 24) ^ (inode >> 16) ^ (inode >> 8) ^ inode) % HASH_SIZE;
}

static int insert_inode(int inode, int pid, char *comm, int seq)
{
	int i = get_hash(inode);
	struct hash_node *hn = hash_lists[i];

	if (!hn) {
		hash_lists[i] = alloc_hash_node(inode, pid, comm, seq);

		return hash_lists[i] ? 0 : -1;
	}

	while (1) {
		if (hn->inode == inode &&
		    hn->pid == pid &&
		    !strcmp(hn->comm, comm)) {
			hn->seq = seq;
			return 0;
		}

		if (!hn->next)
			break;

		hn = hn->next;
	}

	hn->next = alloc_hash_node(inode, pid, comm, seq);

	return hn->next ? 0 : -1;
}

static struct hash_node *get_hash_node(int inode)
{
	int i = get_hash(inode);
	struct hash_node *hn = hash_lists[i];

	while (hn) {
		if (hn->inode == inode)
			return hn;

		hn = hn->next;
	}

	return NULL;
}

static int cleanup_inode_map(int seq)
{
	int i;

	for (i = 0; i < HASH_SIZE; i++) {
		struct hash_node *hn, *head, *tail, *next;

		hn = hash_lists[i];
		if (!hn)
			continue;

		head = tail = NULL;
		while (hn) {
			next = hn->next;
			if (hn->seq < seq) {
				free(hn);
			} else if (!head) {
				head = hn;
				tail = hn;
				hn->next = NULL;
			} else {
				tail->next = hn;
				tail = hn;
				hn->next = NULL;
			}
			hn = next;
		}

		hash_lists[i] = head;
	}
}

static int build_inode_map(int seq)
{
	struct dirent *d;
	DIR *proc;

	proc = opendir("/proc");
	CC(!proc, "opendir\n");

	while ((d = readdir(proc)) != NULL) {
		int fd, len, pid;
		DIR *sub;
		char path[PATH_SIZE];
		char comm[COMM_SIZE];

		if (sscanf(d->d_name, "%d", &pid) != 1 || pid == 1)
			continue;

		snprintf(path, PATH_SIZE, "/proc/%d/comm", pid);
		fd = open(path, O_RDONLY);
		if (fd <= 0) {
			perror("open");
			continue;
		}

		len = read(fd, comm, COMM_SIZE - 1);
		CC(len <= 0, "read\n");

		close(fd);

		comm[len - 1] = '\0';
		if (!target_comm(comm))
			continue;

		snprintf(path, PATH_SIZE, "/proc/%d/fd/", pid);
		sub = opendir(path);
		while ((d = readdir(sub)) != NULL) {
			int inode;
			char link[PATH_SIZE];

			if (sscanf(d->d_name, "%d", &fd) != 1)
				continue;

			snprintf(path, PATH_SIZE, "/proc/%d/fd/%d", pid, fd);
			len = readlink(path, link, PATH_SIZE - 1);
			if (len <= 0)
				continue;

			link[len] = '\0';
			if (sscanf(link, "socket:[%d]", &inode) != 1)
				continue;

			CC(insert_inode(inode, pid, comm, seq), "insert_inode\n");
		}
		closedir(sub);
	}

	closedir(proc);

	cleanup_inode_map(seq);

	return 0;
}

static int send_query(int fd, int family)
{
	struct sockaddr_nl nladdr = {
		.nl_family = AF_NETLINK
	};

	struct
	{
		struct nlmsghdr nlh;
		struct inet_diag_req_v2 idr;
	} req = {
		.nlh = {
			.nlmsg_len = sizeof(req),
			.nlmsg_type = SOCK_DIAG_BY_FAMILY,
			.nlmsg_flags = NLM_F_DUMP | NLM_F_REQUEST,
		},
		.idr = {
			.sdiag_family = family,
			.sdiag_protocol = IPPROTO_TCP,
			.idiag_ext = (1 << (INET_DIAG_INFO - 1)),
			.idiag_states = monitor_states,
		}
	};

	struct iovec iov = {
		.iov_base = &req,
		.iov_len = sizeof(req)
	};

	struct msghdr msg = {
		.msg_name = &nladdr,
		.msg_namelen = sizeof(nladdr),
		.msg_iov = &iov,
		.msg_iovlen = 1
	};

	CC(sendmsg(fd, &msg, 0) < 0, "sendmsg\n");

	return 0;
}

static void pr_bytes(FILE *fd, char *fmt, u64 bytes)
{
	int u = 0;
	double ret;
	char *units[] = {
		"bit",
		"kbit",
		"mbit",
		"gbit",
		"tbit",
	};
	char tmp[PRINT_SIZE];

	bytes *= 8;
	while (bytes > 1048576) {
		bytes /= 1024;
		u++;
	}

	ret = bytes;
	while (ret > 1024) {
		ret /= 1024;
		u++;
	}

	sprintf(tmp, "%0.2f%s", ret, units[u]);
	fprintf(fd, fmt, tmp);
}

static inline void show_records(int iter)
{
	FILE *fd = stdout;
	struct hash_node *hn = show_list, *p95, *p99;
	u64 sent_bytes = 0, recv_bytes = 0, segs_out = 0, retrans = 0;
	int p99_idx = nr_show - (nr_show * 99 / 100);
	int p95_idx = nr_show - (nr_show * 95 / 100);
	int idx = 1;

	if (!nr_show)
		return;

	if (mode & LOG_MODE)
		fd = log_fd;
	else if (!iter)
		system("clear");

	if (mode & TOP_MODE && !iter) {
		time_t t;

		time(&t);
		fprintf(fd, "%s", ctime(&t));
	}

    if (iter == 0 || iter == 1) {
        if (iter > 0)
            fprintf(fd, "%-5s ", "iter");
        fprintf(fd, "%-8s ", "pid");
        fprintf(fd, "%-16s ", "comm");
        fprintf(fd, "%-12s ", "state");
        fprintf(fd, "%-10s ", "ca");
        fprintf(fd, "%-10s ", "recv");
        fprintf(fd, "%-10s ", "sent");
        fprintf(fd, "%-8s ", "retrans");
        fprintf(fd, "%s ", "source -> destination");
        fprintf(fd, "\n");
    }

	while (hn) {
		if (mode & TOP_MODE) {
			segs_out += hn->diff.tcpi_segs_out;
			retrans += hn->diff.tcpi_total_retrans;
			recv_bytes += hn->diff.tcpi_bytes_received;
#if KERNEL_VERSION_AT_LEAST(4, 19)
			sent_bytes += hn->diff.tcpi_bytes_sent;
#endif
			if (idx == p99_idx)
				p99 = hn;
			if (idx == p95_idx)
				p95 = hn;
			if (idx++ > top_show)
				goto next;
		}

        if (iter > 0)
            fprintf(fd, "%-5d ", iter);
		fprintf(fd, "%-8d ", hn->pid);
		fprintf(fd, "%-16s ", hn->comm);
		fprintf(fd, "%-12s ", state_name[hn->ti.tcpi_state]);
		fprintf(fd, "%-10s ", ca_state_name[hn->ti.tcpi_ca_state]);
		pr_bytes(fd, "%-10s ", hn->diff.tcpi_bytes_received);
#if KERNEL_VERSION_AT_LEAST(4, 19)
		pr_bytes(fd, "%-10s ", hn->diff.tcpi_bytes_sent);
#else
		fprintf(fd, "%-10s ", "NaN");
#endif
		fprintf(fd, "%-8.2f ", hn->rr);
		fprintf(fd, "%s:%d -> ", hn->src_ip, hn->src_port);
		fprintf(fd, "%s:%d", hn->dst_ip, hn->dst_port);
		fprintf(fd, "\n");

next:
		hn = hn->plist;
	}

	if (mode & TOP_MODE && !iter) {
		fprintf(fd, "%d connection ", nr_show);
		if (order == RR_ORDER)
			fprintf(fd, "[rr p99=%0.2f%% p95=%0.2f%%] ", p99->rr, p95->rr);
		else if (order == RECV_ORDER) {
			pr_bytes(fd, "[recv p99=%s ", p99->diff.tcpi_bytes_received);
			pr_bytes(fd, "p95=%s] ", p95->diff.tcpi_bytes_received);
		} else if (order == SENT_ORDER) {
#if KERNEL_VERSION_AT_LEAST(4, 19)
			pr_bytes(fd, "[sent p99=%s ", p99->diff.tcpi_bytes_sent);
			pr_bytes(fd, "p95=%s] ", p95->diff.tcpi_bytes_sent);
#else
			fprintf(fd, "[sent p99=%s ", "NaN");
			fprintf(fd, "p95=%s] ", "NaN");
#endif
		}
		fprintf(fd, "rr=%0.2f%% ", (double)retrans * 100 / segs_out);
		pr_bytes(fd, "recv=%s ", recv_bytes);
		pr_bytes(fd, "sent=%s ", sent_bytes);
		fprintf(fd, "\n");
	}

	if (mode & LOG_MODE)
		fflush(log_fd);
}

#define CAL_DIFF(c)	(hn->diff.c = ti->c - hn->ti.c)

static inline void record_tcp_info(struct hash_node *hn, struct tcp_info *ti)
{
	struct hash_node *prev = NULL, *next = show_list;

	hn->rr = 0;
	hn->ti.tcpi_state = ti->tcpi_state;
	CAL_DIFF(tcpi_segs_out);
	CAL_DIFF(tcpi_total_retrans);
	CAL_DIFF(tcpi_bytes_received);
#if KERNEL_VERSION_AT_LEAST(4, 19)
	CAL_DIFF(tcpi_bytes_sent);
#endif

	if (hn->diff.tcpi_segs_out)
		hn->rr = (double)hn->diff.tcpi_total_retrans * 100 / (hn->diff.tcpi_segs_out);

	memcpy(&hn->ti, ti, sizeof(struct tcp_info));

	if (hn->rr < rr_wmark ||
	    hn->diff.tcpi_bytes_received < recv_wmark ||
#if KERNEL_VERSION_AT_LEAST(4, 19)
	    hn->diff.tcpi_bytes_sent < sent_wmark)
#else
		0 < sent_wmark)
#endif
		return;

	while (next) {
		if ((order == RR_ORDER && hn->rr > next->rr) ||
		    (order == RECV_ORDER && hn->diff.tcpi_bytes_received > next->diff.tcpi_bytes_received) ||
#if KERNEL_VERSION_AT_LEAST(4, 19)
		    (order == SENT_ORDER && hn->diff.tcpi_bytes_sent > next->diff.tcpi_bytes_sent))
#else
		    (order == SENT_ORDER && 0))
#endif
			break;
		prev = next;
		next = next->plist;
	}

	if (!prev) {
		hn->plist = show_list;
		show_list = hn;
	} else {
		hn->plist = prev->plist;
		prev->plist = hn;
	}

	nr_show++;
}

static int record_diag(struct inet_diag_msg *res, unsigned int len, int family)
{
	unsigned int rta_len = len - NLMSG_LENGTH(sizeof(*res));
	struct rtattr *attr = (struct rtattr *) (res + 1);

	CC(len < NLMSG_LENGTH(sizeof(*res)), "short response\n");
	CC(res->idiag_family != family, "unexpected family %u\n", res->idiag_family);

	while (RTA_OK(attr, rta_len)) {
		struct hash_node *hn;
		struct tcp_info *ti;

		switch (attr->rta_type) {
		case INET_DIAG_INFO:
			CC(RTA_PAYLOAD(attr) < sizeof(struct tcp_info), "payload too small\n");
			
			ti = (struct tcp_info *)RTA_DATA(attr);

			hn = get_hash_node(res->idiag_inode);
			if (!hn)
				break;

			if (hn->first) {
				hn->src_port = ntohs(res->id.idiag_sport);
				hn->dst_port = ntohs(res->id.idiag_dport);
				CC(!inet_ntop(family, res->id.idiag_src, hn->src_ip, IP_SIZE) ||
				   !inet_ntop(family, res->id.idiag_dst, hn->dst_ip, IP_SIZE),
					    "inet_ntop\n");
				memcpy(&hn->ti, ti, sizeof(struct tcp_info));
				hn->first = 0;
				if (!target_ip(hn))
					hn->disable = 1;
			} else if (!hn->disable)
				record_tcp_info(hn, ti);

			break;
		default:
			break;
		}

		attr = RTA_NEXT(attr, rta_len);
	}
	return 0;
}

static int receive_responses(int fd, int family)
{
	long buf[MSG_SIZE];
	struct sockaddr_nl nladdr;
	struct iovec iov = {
		.iov_base = buf,
		.iov_len = sizeof(buf)
	};
	int flags = MSG_TRUNC;

	while (1) {
		const struct nlmsghdr *h;
		struct msghdr msg = {
			.msg_name = &nladdr,
			.msg_namelen = sizeof(nladdr),
			.msg_iov = &iov,
			.msg_iovlen = 1
		};
		ssize_t ret = recvmsg(fd, &msg, flags);

		CC(ret < 0 || ret > MSG_SIZE * 8, "recvmsg return %llu\n", ret);
		if (ret == 0)
			return 0;

		h = (struct nlmsghdr *) buf;

		CC(nladdr.nl_family != AF_NETLINK, "nl_family wrong\n");

		for (; NLMSG_OK(h, ret); h = NLMSG_NEXT(h, ret)) {
			if (h->nlmsg_type == NLMSG_DONE)
				return 0;

			CC(h->nlmsg_type == NLMSG_ERROR, "msg error\n");
			CC(h->nlmsg_type != SOCK_DIAG_BY_FAMILY,
				"unexpected msg type %u\n", h->nlmsg_type);
			CC(record_diag(NLMSG_DATA(h), h->nlmsg_len, family),
				"record_diag\n");
		}
	}

	return 0;
}

int main(int argc, char **argv)
{
	int opt, fd, seq = 1;
	u64 period_delay_ms = 1000000;
    int iter_num = -1, iter = 0;

	while ((opt = getopt(argc, argv, ":d:m:o:l:n:")) != -1) {
		switch (opt) {
		case 'd':
			period_delay_ms = atoi(optarg) * 1000;
			break;
		case 'm':
			if (sscanf(optarg, "top=%d", &top_show) == 1)
				mode |= TOP_MODE;
			else if (sscanf(optarg, "rr>%d", &rr_wmark) == 1)
				mode |= FILTER_MODE;
			else if (sscanf(optarg, "recv>%llu", &recv_wmark) == 1)
				mode |= FILTER_MODE;
			else if (sscanf(optarg, "sent>%llu", &sent_wmark) == 1)
				mode |= FILTER_MODE;
			else if (setup_target_comm(optarg) && setup_target_ip(optarg))
				goto help;
			break;
		case 'o':
			if (!strcmp(optarg, "rr"))
				order = RR_ORDER;
			else if (!strcmp(optarg, "recv"))
				order = RECV_ORDER;
			else if (!strcmp(optarg, "sent"))
				order = SENT_ORDER;
			break;
		case 'l':
			strncpy(log_path, optarg, PATH_SIZE);
			log_fd = fopen(log_path, "w+");
			CC(!log_fd, "fopen\n");
			mode |= LOG_MODE;
			break;
        case 'n':
            iter_num = atoi(optarg);
			break;
		default:
help:
			printf("Usage: %s [-d period_delay_ms] [-m mode] [-o order] [-l log_path] [-n iter_num]\n", argv[0]);
			printf("mode: \n\tcomm=X  -- show specified task only\
				      \n\tcomm!=X -- show all but specified task only\
				      \n\tsrc!=X  -- show all but specified source address\
				      \n\tsrc=X   -- show specified source address only\
				      \n\tdst!=X  -- show all but specified destination address\
				      \n\tdst=X   -- show specified destination address only\
				      \n\ttop=X   -- show in top X mode\
				      \n\trr>X    -- show retranse-rate higher than X only\
				      \n\trecv>X  -- show received bytes higher than X only\
				      \n\tsent>X  -- show sent bytes higher than X only\n");
			printf("order:\n\trr   -- show in retranse-rate order\
				      \n\trecv -- show in received order\
				      \n\tsent -- show in sent order\n");
			exit(EXIT_FAILURE);
		}
	}

	fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_SOCK_DIAG);
	CC(fd < 0, "socket\n");

	while (iter_num < 0 || iter <= iter_num) {
		nr_show = 0;
		show_list = NULL;

		if (build_inode_map(seq++)		||
		    send_query(fd, AF_INET)		||
		    receive_responses(fd, AF_INET)	||
		    send_query(fd, AF_INET6)		||
		    receive_responses(fd, AF_INET6))
			break;

		show_records(iter);

        if (iter_num >= 0)
            iter++;

		usleep(period_delay_ms);
	}

	close(fd);
    if (mode & LOG_MODE)
        fclose(log_fd);
	return 0;
}

